// SPDX-FileCopyrightText: Copyright (c) 2025 Cisco and/or its affiliates.
// SPDX-License-Identifier: Apache-2.0

pub mod pubsub {
    pub mod v1 {
        include!("gen/pubsub.proto.v1.rs");
    }
}
